#define G_LOG_DOMAIN "ap-sheet"

#include "sheet.h"

#include <glib.h>
#include <glib-object.h>
#include <gio/gio.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

typedef struct _APSheetPrivate
{
    xmlDoc *doc;
    gchar *filename;

    GHashTable *sections;
} APSheetPrivate;

struct _APSheet
{
    GObject parent_instance;
};

G_DEFINE_TYPE_WITH_PRIVATE (APSheet, ap_sheet, G_TYPE_OBJECT)

enum
{
    PROP_FILE,
    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };

gboolean
ap_sheet_document_exists (APSheet *self)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    if (priv->doc != NULL)
        return TRUE;
    else
        return FALSE;
}

void
ap_sheet_create_document (APSheet *self)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    if (priv->doc == NULL)
    {
        xmlNode *root_node = NULL;

        priv->doc = xmlNewDoc (BAD_CAST "1.0");
        root_node = xmlNewNode (NULL, BAD_CAST "sheet");
        xmlDocSetRootElement (priv->doc, root_node);
    }
}

void
ap_sheet_create_document_if_no_exists (APSheet *self)
{
    if (!ap_sheet_document_exists (self))
    {
        ap_sheet_create_document (self);
    }
}

void
ap_sheet_save_to_file (APSheet     *self,
                       const gchar *filename)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);
    int status;

    if (filename == NULL)
    {
        g_warning ("Can not provide NULL for filename in ap_sheet_save_to_file (), no file to save to");
        return;
    }

    g_debug ("Loading sheet from file, this will overwrite the current sheet");

    if (priv->doc != NULL)
    {
        xmlFreeDoc (priv->doc);
    }

    status = xmlSaveFormatFileEnc (filename, priv->doc, "UTF-8", 1);

    if (status < 0)
    {
        g_warning ("Unable to save to \"%s\"", filename);
    }
}

void
ap_sheet_save_to_current_file (APSheet *self)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    ap_sheet_save_to_file (self, priv->filename);
}

void
ap_sheet_load_from_file (APSheet     *self,
                         const gchar *filename)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    if (filename == NULL)
    {
        g_warning ("Can not provide NULL for filename in ap_sheet_load_from_file (), no file loaded");
        return;
    }

    g_debug ("Loading sheet from file, this will overwrite the current sheet");

    if (priv->doc != NULL)
    {
        xmlFreeDoc (priv->doc);
    }

    priv->doc = xmlReadFile (filename, "UTF-8", 0);

    if (priv->doc == NULL)
    {
        g_warning ("Could not parse file \"%s\"", filename);
    }
}

xmlDoc *
ap_sheet_get_doc (APSheet *self)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    g_return_val_if_fail (priv->doc != NULL, NULL);

    return priv->doc;
}
/*
void
ap_sheet_print_sections (APSheet *self)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    g_hash_table_foreach (priv->sections,
                          g_print,
                          NULL);
}
*/
xmlNode *
ap_sheet_get_section_by_name (APSheet     *self,
                              const gchar *section_name)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    return g_hash_table_lookup (priv->sections, section_name);
}

xmlNode *
ap_sheet_add_section (APSheet     *self,
                      const gchar *section_name)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);
    xmlNode *section;

    ap_sheet_create_document_if_no_exists (self);

    section = ap_sheet_get_section_by_name (self, section_name);

    if (section == NULL)
    {
        xmlChar *name;
        xmlNode *root;

        name = xmlCharStrdup (section_name);
        root = xmlDocGetRootElement (priv->doc);

        section = xmlNewChild (root, NULL, name, NULL);

        g_hash_table_insert (priv->sections, g_strdup (section_name), section);
    }

    return section;
}

xmlNode *
ap_sheet_add_section_with (APSheet     *self,
                           const gchar *section_name,
                           xmlNode     *node)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);
    xmlNode *root;
    xmlNode *section;

    ap_sheet_create_document_if_no_exists (self);

    if (section_name == NULL)
    {
        // get node name


        if (section_name == NULL)
        {
            g_warning ("Could not find an existing name in node, can not add a section with no name. No node added.");
            return NULL;
        }
    }

    section = ap_sheet_get_section_by_name (self, section_name);

    if (section != NULL)
    {
        g_warning ("Can not add node \"%s\", as a node with the same name already exists", section_name);
        return NULL;
    }

    root = xmlDocGetRootElement (priv->doc);

    return xmlAddChild (root, node);
}

xmlNode *
ap_sheet_section_add (APSheet     *self,
                      const gchar *section_name,
                      xmlNode     *node)
{
    xmlNode *section;

    section = ap_sheet_add_section (self, section_name);

    return xmlAddChild (section, node);
}

static void
ap_sheet_set_property (GObject      *object,
                       guint         property_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
    APSheet *self = AP_SHEET (object);

    switch (property_id)
    {
        case PROP_FILE:
            ap_sheet_load_from_file (self, g_value_dup_string (value));
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_sheet_get_property (GObject      *object,
                       guint         property_id,
                       GValue       *value,
                       GParamSpec   *pspec)
{
    APSheet *self = AP_SHEET (object);
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    switch (property_id)
    {
        case PROP_FILE:
            g_value_set_string (value, priv->filename);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

/*
 * boilerplate_window_constructed:
 * @object: An instance of #BoilerplateWindow
 *
 * The constructed virtual function sets first-run settings.
 */
static void
ap_sheet_constructed (GObject *object)
{
    APSheet *self = AP_SHEET (object);

    ap_sheet_create_document (self);

    G_OBJECT_CLASS (ap_sheet_parent_class)->constructed (object);
}

/*
 * ap_sheet_finalize:
 * @object: An instance of #APSheet
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
ap_sheet_finalize (GObject *object)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (AP_SHEET (object));

    if (priv->doc != NULL)
    {
        xmlFreeDoc (priv->doc);
    }

    xmlCleanupParser ();

    G_OBJECT_CLASS (ap_sheet_parent_class)->finalize (object);
}

/*
 * ap_sheet_init:
 * @self: An instance of #APSheet
 *
 * Initializes various aspects of the sheet.
 */
static void
ap_sheet_init (APSheet *self)
{
    APSheetPrivate *priv = ap_sheet_get_instance_private (self);

    priv->sections = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
}

/*
 * ap_sheet_class_init:
 * @klass: An instance of #APSheetClass
 *
 * Initializes the various virtural functions in #APSheetClass.
 */
static void
ap_sheet_class_init (APSheetClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->constructed = ap_sheet_constructed;
    object_class->finalize = ap_sheet_finalize;

    object_class->set_property = ap_sheet_set_property;
    object_class->get_property = ap_sheet_get_property;

    obj_properties[PROP_FILE] =
        g_param_spec_string ("file",
                             "file",
                             "Name of the file to load and display from.",
                             NULL  /* default value */,
                             G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

    g_object_class_install_properties (object_class,
                                       N_PROPERTIES,
                                       obj_properties);
}

/**
 * ap_sheet_new:
 *
 * Creates a new #APSheet, a sub-object of #GtkWindow.
 *
 * ## Usage
 * |[
 * APSheet *sheet;
 *
 * sheet = ap_sheet_new ();
 * ]|
 *
 * Returns: #APSheet; a sub-object of #GObject
 */
APSheet *
ap_sheet_new (void)
{
    return g_object_new (AP_TYPE_SHEET,
                         NULL);
}

/**
 * ap_sheet_new:
 *
 * Creates a new #APSheet, a sub-object of #GtkWindow.
 *
 * ## Usage
 * |[
 * APSheet *sheet;
 *
 * sheet = ap_sheet_from_file ("./my-sheet.xml");
 * ]|
 *
 * Returns: #APSheet; a sub-object of #GObject
 */
APSheet *
ap_sheet_from_file (const gchar *file)
{
    return g_object_new (AP_TYPE_SHEET,
						 "file", file,
                         NULL);
}


